local obj_get_nearest_object_with_behavior_id, smlua_model_util_get_id, network_player_from_global_index, obj_mark_for_deletion,
obj_check_overlap_with_hitbox_params, obj_check_hitbox_overlap, nearest_mario_state_to_object = 
obj_get_nearest_object_with_behavior_id, smlua_model_util_get_id, network_player_from_global_index, obj_mark_for_deletion,
obj_check_overlap_with_hitbox_params, obj_check_hitbox_overlap, nearest_mario_state_to_object

------------------------
--- GENERAL HITBOXES ---
------------------------

-- Functions
function set_hitbox_pos_and_angle_to_obj_pos_angle(o)
    o.oHitboxPosX = o.oPosX
    o.oHitboxPosY = o.oPosY
    o.oHitboxPosZ = o.oPosZ
    o.oHitboxYaw = o.oMoveAngleYaw
    o.oHitboxPitch = o.oMoveAnglePitch
end

-- CONSTANTS

E_model_hitcylinder = smlua_model_util_get_id('hitcylinder_geo')

HITBOX_SHOT_WEAK = 1
HITBOX_SHOT_MEDIUM = 2
HITBOX_SHOT_STRONG = 3

HITBOX_MELEE_LIGHT_SLASH = 100
HITBOX_MELEE_MEDIUM_SLASH = 102
HITBOX_MELEE_HEAVY_SLASH = 103
HITBOX_MELEE_AIR_SLASH = 104
HITBOX_MELEE_KUENZAN = 105
HITBOX_MELEE_SHIPPUGA = 106
HITBOX_MELEE_RAIKOUSEN = 107
HITBOX_MELEE_WALL_SLIDE_SLASH = 108
HITBOX_MELEE_HYOURETSUZAN = 109
HITBOX_MELEE_RYUENJIN = 110
HITBOX_MELEE_SENTSUIZAN = 111
HITBOX_MELEE_RAIKOUSEN_LIGHTNING = 120

HITBOX_BURST_LEG_NAPALM = 1000
HITBOX_BURST_GAEA_GIGAATTACK = 1001

local HITBOX_NO_HIT = 0
local HITBOX_HIT_DEATH = 1
local HITBOX_HIT_NO_DEATH = 2

------------------------------
--- MAIN HITBOX DEFINITION ---
------------------------------

local HITBOX_GROUP_PROJECTILE = {
    [HITBOX_SHOT_WEAK] = true,
    [HITBOX_SHOT_MEDIUM] = true,
    [HITBOX_SHOT_STRONG] = true,
}

-- Sabers, Rapiers, etc...
local HITBOX_GROUP_MELEE = {
    [HITBOX_MELEE_LIGHT_SLASH] = true,
    [HITBOX_MELEE_MEDIUM_SLASH] = true,
    [HITBOX_MELEE_HEAVY_SLASH] = true,
    [HITBOX_MELEE_AIR_SLASH] = true,
    [HITBOX_MELEE_KUENZAN] = true,
    [HITBOX_MELEE_SHIPPUGA] = true,
    [HITBOX_MELEE_RAIKOUSEN] = true,
    [HITBOX_MELEE_WALL_SLIDE_SLASH] = true,
    [HITBOX_MELEE_HYOURETSUZAN] = true,
    [HITBOX_MELEE_RYUENJIN] = true,
    [HITBOX_MELEE_SENTSUIZAN] = true,
    [HITBOX_MELEE_RAIKOUSEN_LIGHTNING] = true,
}

-- Explosions and Giga Attacks
local HITBOX_GROUP_BURST = {
    [HITBOX_BURST_LEG_NAPALM] = true,
    [HITBOX_BURST_GAEA_GIGAATTACK] = true,
}

local hitbox_type_parent = {}

function map_hitbox_type_to_valid_parents()
    hitbox_type_parent = {
        [HITBOX_SHOT_WEAK] = id_bhvShot,
        [HITBOX_SHOT_MEDIUM] = id_bhvShot,
        [HITBOX_SHOT_STRONG] = id_bhvShot,
        [HITBOX_MELEE_LIGHT_SLASH] = id_bhvSaber,
        [HITBOX_MELEE_MEDIUM_SLASH] = id_bhvSaber,
        [HITBOX_MELEE_HEAVY_SLASH] = id_bhvSaber,
        [HITBOX_MELEE_AIR_SLASH] = id_bhvSaber,
        [HITBOX_MELEE_KUENZAN] = id_bhvSaber,
        [HITBOX_MELEE_SHIPPUGA] = id_bhvSaber,
        [HITBOX_MELEE_RAIKOUSEN] = id_bhvSaber,
        [HITBOX_MELEE_WALL_SLIDE_SLASH] = id_bhvSaber,
        [HITBOX_MELEE_HYOURETSUZAN] = id_bhvSaber,
        [HITBOX_MELEE_RYUENJIN] = id_bhvSaber,
        [HITBOX_MELEE_SENTSUIZAN] = id_bhvSaber,
        [HITBOX_MELEE_RAIKOUSEN_LIGHTNING] = id_bhvRaikousenLightningHitboxSpawner,
        [HITBOX_BURST_LEG_NAPALM] = id_bhvLegNapalmExplosion,
        [HITBOX_BURST_GAEA_GIGAATTACK] = id_bhvGaeaGigaAttack,
    }
end

--------------
--- EXTRAS ---
--------------

local HITBOX_GROUP_NPC_DESTROY = {
    [HITBOX_BURST_LEG_NAPALM] = true,
}

function assign_id_to_hitbox(m)
    local mmx = gMmxStates[m.playerIndex]
    local i = 0
    while mmx.active_hitboxes[i] do
      i = i + 1
    end
    mmx.active_hitboxes[i] = {}
    network_send(true, { type = 'hitbox_id_sync', id = i, m = gNetworkPlayers[m.playerIndex].globalIndex, remove = 0 })
    return i
end

-----------------------------------------------------------------
-----------------------------------------------------------------
----- INTERACTIONS ----------------------------------------------
-----------------------------------------------------------------
-----------------------------------------------------------------

function spawn_hit_fx(o, hitbox_pos, clank, hitbox_type)
    local posx = hitbox_pos.x - (hitbox_pos.x - o.oPosX) / 2
    local posy = hitbox_pos.y - (hitbox_pos.y - o.oPosY) / 2
    local posz = hitbox_pos.z - (hitbox_pos.z - o.oPosZ) / 2
    local hit_model
    if HITBOX_GROUP_MELEE[hitbox_type] == true then
        hit_model = E_model_saber_dmg
        if (clank == 1) then
            mmx_play_sound(SOUND_CLANK_HIT, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}, 0.4)  
        else
            mmx_play_sound(SOUND_ZERO_SABER_HIT, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}, 0.5)
        end
    else
        hit_model = E_model_shot_dmg
        if (clank == 1) then
            mmx_play_sound(SOUND_CLANK_HIT, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}, 0.4)  
        else
            mmx_play_sound(SOUND_SHOT_HIT, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}, 0.5)
        end
    end
    spawn_non_sync_object(id_bhvWallspark, hit_model, posx , posy, posz, function (obj)
        obj.oMoveAngleYaw = calculate_yaw(hitbox_pos, {x = o.oPosX, y = o.oPosY, z = o.oPosZ})
    end)
end

function spawn_destroy_explosion(o, size, hitbox_type)
    spawn_non_sync_object(id_bhvWhitePuffSmoke2, E_MODEL_EXPLOSION, o.oPosX , o.oPosY + 20 * size, o.oPosZ , function(obj)
        obj_scale(obj, size)
    end)
    if HITBOX_GROUP_MELEE[hitbox_type] == true then
        mmx_play_sound(SOUND_ZERO_SABER_DESTROY, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}, 0.9)
    else
        mmx_play_sound(SOUND_SHOT_DESTROY, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}, 0.6)
    end 
end

--------
--NPCs--
--------

function hitbox_interact_bobomb_buddy(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not HITBOX_GROUP_NPC_DESTROY[hitbox_type] then
        return HITBOX_NO_HIT
    end
    spawn_non_sync_object(id_bhvExplosion, E_MODEL_EXPLOSION, o.oPosX, o.oPosY, o.oPosZ, nil)
    obj_mark_for_deletion(o)
end

function hitbox_interact_srgeneral_npc(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not HITBOX_GROUP_NPC_DESTROY[hitbox_type] then
        return HITBOX_NO_HIT
    end
    spawn_mist_particles()
    play_sound(SOUND_OBJ_DEFAULT_DEATH, o.header.gfx.cameraToObject)
    obj_mark_for_deletion(o)
end

function hitbox_interact_general_npc(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not HITBOX_GROUP_NPC_DESTROY[hitbox_type] then
        return HITBOX_NO_HIT
    end
    spawn_mist_particles()
    play_sound(SOUND_OBJ_DEFAULT_DEATH, o.header.gfx.cameraToObject)
    obj_mark_for_deletion(o)
end

function hitbox_interact_mips(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not HITBOX_GROUP_NPC_DESTROY[hitbox_type] then
        return HITBOX_NO_HIT
    end
    bhv_spawn_star_no_level_exit(o, o.oHitboxCanHitTimer + 3, true)
    spawn_mist_particles()
    play_sound(SOUND_OBJ_SWOOP_DEATH, o.header.gfx.cameraToObject)
    obj_mark_for_deletion(o)
end

function hitbox_interact_ukiki_cage(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not HITBOX_GROUP_NPC_DESTROY[hitbox_type] then
        return HITBOX_NO_HIT
    end
    if (o.oAction == UKIKI_CAGE_ACT_WAIT_FOR_UKIKI) then
        o.oAction = UKIKI_CAGE_ACT_FALL
    end
end


-------------------
--GENERAL ENEMIES--
-------------------

function hitbox_interact_srnormal(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED) then
        return HITBOX_NO_HIT
    end
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    o.oInteractStatus = ATTACK_PUNCH | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
    return HITBOX_HIT_DEATH
end

function hitbox_interact_normal(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED) then
        return HITBOX_NO_HIT
    end
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    o.oInteractStatus = ATTACK_PUNCH | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
    return HITBOX_HIT_DEATH
end

function hitbox_interact_chuckya(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        o.usingObj = nil
        obj_mark_for_deletion(o)
        obj_spawn_loot_yellow_coins(o, 5, 20.0)
        spawn_mist_particles_with_sound(SOUND_OBJ_CHUCKYA_DEATH)
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_enemygeneral(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oMmxHealth <= 0) or (o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED) then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_GROUND_POUND_OR_TWIRL | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_bigenemygeneral(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oMmxHealth <= 0) or (o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED) then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        o.oMmxHealth = o.oMmxInitialHealth
        o.oInteractStatus = ATTACK_GROUND_POUND_OR_TWIRL | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

-----------------------
--BOB-OMB BATTLEFIELD--
-----------------------

-- Koopa

function hitbox_interact_koopa(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oMmxHealth <= 0) or (o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED)
    or o.oKoopaMovementType >= KOOPA_BP_KOOPA_THE_QUICK_BASE then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oMmxHealth = o.oMmxInitialHealth
        o.oInteractStatus = ATTACK_GROUND_POUND_OR_TWIRL | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

-- Bob-omb

local HITBOX_GROUP_PUSH_BOBOMB = {
    [HITBOX_SHOT_WEAK] = true,
    [HITBOX_MELEE_LIGHT_SLASH] = true,
    [HITBOX_MELEE_MEDIUM_SLASH] = true,
    [HITBOX_MELEE_AIR_SLASH] = true,
    [HITBOX_MELEE_KUENZAN] = true,
}

function hitbox_interact_bobomb(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == BOBOMB_ACT_EXPLODE then
        return HITBOX_NO_HIT
    end
    --o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    --spawn_destroy_explosion(o, 1, hitbox_type)
    if HITBOX_GROUP_PUSH_BOBOMB[hitbox_type] then
        spawn_hit_fx(o, hitbox_pos, 1, hitbox_type)
        local angleDiff = calculate_yaw(hitbox_pos, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}) - moveAngle
        if angleDiff > 32768 then
            angleDiff = -2 * 32768 + angleDiff
        elseif angleDiff < -32768 then
            angleDiff = 2 * 32768 + angleDiff
        end
        o.oMoveAngleYaw = moveAngle + angleDiff / 2
        o.oForwardVel = 25.0
        o.oVelY = 20.0
        o.oAction = BOBOMB_ACT_LAUNCHED
    else
        spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oAction = BOBOMB_ACT_EXPLODE
    end
    return HITBOX_HIT_NO_DEATH
end

-- Chain Chomp

function hitbox_interact_chainchomp(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 2, hitbox_type)
        spawn_mist_particles_with_sound(SOUND_GENERAL_WALL_EXPLOSION)
        spawn_triangle_break_particles(20, 138, 3, 4);
        obj_mark_for_deletion(o)
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_waterbomb(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAnglex)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oAction = WATER_BOMB_ACT_EXPLODE
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_kingbobomb(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oFlags & OBJ_FLAG_HOLDABLE) == 0 or o.oAction == 8 or o.oAction == 7 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.5, hitbox_type)
        o.oMoveFlags = OBJ_MOVE_LANDED
        o.oHealth = 0
        o.usingObj = nil
        o.oForwardVel = 0
        o.oVelY = 0
        o.oAction = 7
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_breakablebox(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        o.oInteractStatus =  ATTACK_PUNCH | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_breakableboxsmall(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_KICK_OR_TRIP | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | INT_STATUS_STOP_RIDING
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_exclamationbox(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction ~= 2 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_KICK_OR_TRIP | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

--------------------
--WHOMP'S FORTRESS--
--------------------

function hitbox_interact_breakablewall(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oMmxHealth <= 0 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.5, hitbox_type)
        spawn_mist_particles_with_sound(SOUND_GENERAL_WALL_EXPLOSION)
        --obj_explode_and_spawn_coins(80, 0)
        obj_mark_for_deletion(o)
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_bridge(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not (o.oFaceAnglePitch > -0x4000) then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        o.oAction = 2
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_whomp(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        o.oNumLootCoins = 5
        obj_spawn_loot_yellow_coins(o, 5, 20.0)
        o.oAction = 8
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_whompking(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == 8 or o.oAction == 0 or o.oAction == 9 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 2, hitbox_type)
        play_sound(SOUND_OBJ2_WHOMP_SOUND_SHORT, o.header.gfx.cameraToObject)
        play_sound(SOUND_OBJ_KING_WHOMP_DEATH, o.header.gfx.cameraToObject)
        o.oHealth = 0
        o.oMoveFlags = OBJ_MOVE_ON_GROUND
        o.oForwardVel = 0
        o.oAction = 8
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_piranhaplant(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == PIRANHA_PLANT_ACT_SHRINK_AND_DIE 
    or o.oAction == PIRANHA_PLANT_ACT_ATTACKED or o.oAction == PIRANHA_PLANT_ACT_WAIT_TO_RESPAWN then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        stop_secondary_music(50)
        o.oAction = PIRANHA_PLANT_ACT_ATTACKED
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_bulletbill(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == 3 or o.oAction == 4 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        o.oAction = 3
        spawn_non_sync_object(id_bhvExplosion, E_MODEL_EXPLOSION, o.oPosX, o.oPosY, o.oPosZ, nil)
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

-------------------
--JOLLY ROGER BAY--
-------------------

function hitbox_interact_hiddenobject(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction ~= 1 then
        return HITBOX_NO_HIT
    end
    return hitbox_interact_breakablebox(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
end

----------------------
--COOL COOL MOUNTAIN--
----------------------

function hitbox_interact_mrblizzard(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == MR_BLIZZARD_ACT_DEATH then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oAction = MR_BLIZZARD_ACT_DEATH
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

----------------------
--BIG BOO'S HUNT------
----------------------
function hitbox_interact_boo(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if (o.oMmxHealth <= 0) or (o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED) then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oAction = 3
        if o.oHealth > 1 then
            o.oMmxHealth = o.oMmxInitialHealth
        else
            return HITBOX_HIT_DEATH
        end
    end
    return HITBOX_HIT_NO_DEATH
end

----------------------
--SHIFTING SAND LAND--
----------------------
function hitbox_interact_klepto(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oInteractStatus == ATTACK_FROM_ABOVE | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | INT_STATUS_STOP_RIDING then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_FROM_ABOVE | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | INT_STATUS_STOP_RIDING
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_jumping_box(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = INT_STATUS_STOP_RIDING
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_eyerokhand(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == EYEROK_HAND_ACT_DIE or o.oAction == EYEROK_HAND_ACT_DEAD then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = INT_STATUS_STOP_RIDING
        o.parentObj.oEyerokBossNumHands = o.parentObj.oEyerokBossNumHands - 1
        o.oAction = EYEROK_HAND_ACT_DIE
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

--------------------
--LETHAL LAVA LAND--
--------------------
function hitbox_interact_mri(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == 3 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oAction = 3
        o.oTimer = 104
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_bully(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMoveAngleYaw = moveAngle
    if HITBOX_GROUP_MELEE[hitbox_type] and hitbox_type ~= HITBOX_MELEE_RYUENJIN then
        local angleDiff = calculate_yaw(hitbox_pos, {x = o.oPosX, y = o.oPosY, z = o.oPosZ}) - moveAngle
        if angleDiff > 32768 then
            angleDiff = -2 * 32768 + angleDiff
        elseif angleDiff < -32768 then
            angleDiff = 2 * 32768 + angleDiff
        end
        o.oMoveAngleYaw = moveAngle + angleDiff / 2
    end
    spawn_hit_fx(o, hitbox_pos, 1, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if HITBOX_GROUP_PROJECTILE[hitbox_type] then
        if hitbox_dmg < 1 then
            o.oInteractStatus = INT_STATUS_INTERACTED
            o.oForwardVel = 15
        elseif hitbox_dmg >= 1 and hitbox_dmg < 4 then
            o.oInteractStatus = ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
            o.oForwardVel = 30
        else
            o.oInteractStatus = ATTACK_GROUND_POUND_OR_TWIRL | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
            o.oForwardVel = 50
        end
    elseif HITBOX_GROUP_MELEE[hitbox_type] then
        o.oInteractStatus = ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        o.oForwardVel = 35
    end
    return HITBOX_HIT_NO_DEATH
end

---------------------
---TINY HUGE ISLAND--
---------------------

function hitbox_interact_spiny(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        spawn_mist_particles()
        play_sound(SOUND_OBJ_DEFAULT_DEATH, o.header.gfx.cameraToObject)
        o.oNumLootCoins = 1
        obj_spawn_loot_yellow_coins(o, 1, 20.0);
        obj_mark_for_deletion(o)
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_firepiranhaplant(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == FIRE_PIRANHA_PLANT_ACT_HIDE then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_wigglerhead(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oIntangibleTimer < 0 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_FROM_ABOVE | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | INT_STATUS_STOP_RIDING
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

------------------------
---TALL TALL MOUNTAIN---
------------------------

function hitbox_interact_monty_mole(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == MONTY_MOLE_ACT_SELECT_HOLE and 
    o.oInteractStatus == ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1, hitbox_type)
        o.oInteractStatus = ATTACK_FAST_ATTACK | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
        o.oMmxHealth = o.oMmxInitialHealth
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

----------------------------
--BOWSER IN THE DARK WORLD--
----------------------------

function hitbox_interact_bowser(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if o.oAction == 19 or o.oAction == 5 or o.oAction == 6 or o.oAction == 4 or o.oAction == 1 then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        o.oHealth = o.oHealth - 1
        spawn_destroy_explosion(o, 1.8, hitbox_type)
        if o.oHealth <= 0 then
            o.oAction = 4
        else
            o.oMmxHealth = o.oMmxInitialHealth
            o.oAction = 12
        end
        o.oMoveFlags = 0
        o.oForwardVel = 0
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_bowser_bomb(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_non_sync_object(id_bhvExplosion, E_MODEL_EXPLOSION, o.oPosX, o.oPosY, o.oPosZ, nil)
        obj_mark_for_deletion(o)
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end


---------
--OTHER--
---------

function hitbox_interact_srbreakable(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_hit_fx(o, hitbox_pos, 0, hitbox_type)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 1.2, hitbox_type)
        spawn_triangle_break_particles(20, 138, 3, 4);
        obj_mark_for_deletion(o)
        return HITBOX_HIT_DEATH
    end
    return HITBOX_HIT_NO_DEATH
end

function hitbox_interact_srexplosion(o, hitbox_id, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, moveAngle)
    if not HITBOX_GROUP_BURST[hitbox_type] then
        return HITBOX_NO_HIT
    end
    o.oMmxHealth = o.oMmxHealth - hitbox_dmg
    spawn_shot_fx(o, hitbox_pos, 0)
    local mmx = gMmxStates[playerIndex]
    if mmx.active_hitboxes[hitbox_id] ~= nil and not mmx.active_hitboxes[hitbox_id][o.oSyncID] then
        mmx.active_hitboxes[hitbox_id][o.oSyncID] = true
    end
    if (o.oMmxHealth <= 0) then
        spawn_destroy_explosion(o, 2, hitbox_type)
        spawn_mist_particles_with_sound(SOUND_GENERAL_WALL_EXPLOSION)
        spawn_triangle_break_particles(20, 138, 3, 4);
        obj_mark_for_deletion(o)
    end
    return true
end

-----------------------------------------------------------------
-----------------------------------------------------------------
----- INTERACTIONS HANDLING -------------------------------------
-----------------------------------------------------------------
-----------------------------------------------------------------

local loadedinteractablebhvs = {}

hitboxhurtableenemy = {
    -- NPCs
    [id_bhvBobombBuddy] = hitbox_interact_bobomb_buddy,
    [id_bhvBobombBuddyOpensCannon] = hitbox_interact_bobomb_buddy,
    [id_bhvToadMessage] = hitbox_interact_general_npc,
    [id_bhvYoshi] = hitbox_interact_general_npc,
    [id_bhvUkiki] = hitbox_interact_general_npc,
    [id_bhvHoot] = hitbox_interact_general_npc,
    [id_bhvPenguinBaby] = hitbox_interact_general_npc,
    [id_bhvSmallPenguin] = hitbox_interact_general_npc,
    [id_bhvSLWalkingPenguin] = hitbox_interact_general_npc,
    [id_bhvTuxiesMother] = hitbox_interact_general_npc,
    [id_bhvRacingPenguin] = hitbox_interact_general_npc,
    [id_bhvMips] = hitbox_interact_mips,
    [id_bhvUkikiCage] = hitbox_interact_ukiki_cage,
    -- BOB
    [id_bhvGoomba] = hitbox_interact_enemygeneral, 
    [id_bhvKoopa] = hitbox_interact_koopa,
    [id_bhvBobomb] = hitbox_interact_bobomb,
    [id_bhvChainChomp] = hitbox_interact_chainchomp,
    [id_bhvChainChompGate] = hitbox_interact_chainchomp,
    [id_bhvWaterBomb] = hitbox_interact_waterbomb,
    [id_bhvKingBobomb] = hitbox_interact_kingbobomb,
    [id_bhvBreakableBox] = hitbox_interact_breakablebox,
    [id_bhvBreakableBoxSmall] = hitbox_interact_breakableboxsmall,
    [id_bhvExclamationBox] = hitbox_interact_exclamationbox,
    [id_bhvBowlingBall] = hitbox_interact_chainchomp,
    [id_bhvPitBowlingBall] = hitbox_interact_chainchomp,
    [id_bhvFreeBowlingBall] = hitbox_interact_chainchomp,
    -- WF
    [id_bhvKickableBoard] = hitbox_interact_bridge,
    [id_bhvWfBreakableWallLeft] = hitbox_interact_breakablewall,
    [id_bhvWfBreakableWallRight] = hitbox_interact_breakablewall,
    [id_bhvSmallWhomp] = hitbox_interact_whomp,
    [id_bhvWhompKingBoss] = hitbox_interact_whompking,
    [id_bhvPiranhaPlant] = hitbox_interact_piranhaplant,
    [id_bhvBulletBill] = hitbox_interact_bulletbill,
    [id_bhvTowerDoor] = hitbox_interact_breakablebox,
    -- CCM
    [id_bhvSpindrift] = hitbox_interact_enemygeneral,
    [id_bhvMrBlizzard] = hitbox_interact_mrblizzard,
    -- JRB
    [id_bhvHiddenObject] = hitbox_interact_hiddenobject,
    -- BBH
    [id_bhvBoo] = hitbox_interact_boo,
    [id_bhvGhostHuntBoo] = hitbox_interact_boo,
    [id_bhvFlyingBookend] = hitbox_interact_enemygeneral,
    [id_bhvHauntedChair] = hitbox_interact_enemygeneral,
    [id_bhvBooWithCage] = hitbox_interact_boo,
    [id_bhvBookSwitch] = hitbox_interact_normal,
    [id_bhvMerryGoRoundBoo] = hitbox_interact_boo,
    [id_bhvBalconyBigBoo] = hitbox_interact_boo,
    [id_bhvGhostHuntBigBoo] = hitbox_interact_boo,
    [id_bhvMerryGoRoundBigBoo] = hitbox_interact_boo,
    [id_bhvMadPiano] = hitbox_interact_chainchomp,
    -- LLL
    [id_bhvMrI] = hitbox_interact_mri,
    [id_bhvSmallBully] = hitbox_interact_bully,
    [id_bhvBigBully] = hitbox_interact_bully,
    [id_bhvBigBullyWithMinions] = hitbox_interact_bully,
    -- SSL
    [id_bhvPokey] = hitbox_interact_enemygeneral,
    [id_bhvPokeyBodyPart] = hitbox_interact_enemygeneral,
    [id_bhvKlepto] = hitbox_interact_klepto,
    [id_bhvFlyGuy] = hitbox_interact_enemygeneral,
    [id_bhvEyerokHand] = hitbox_interact_eyerokhand,
    [id_bhvJumpingBox] = hitbox_interact_jumping_box,
    -- HMC
    [id_bhvScuttlebug] = hitbox_interact_enemygeneral,
    [id_bhvSnufit] = hitbox_interact_enemygeneral,
    [id_bhvSwoop] = hitbox_interact_enemygeneral,
    [id_bhvBigBoulder] = hitbox_interact_chainchomp,
    -- THI
    [id_bhvFirePiranhaPlant] = hitbox_interact_firepiranhaplant,
    [id_bhvChuckya] = hitbox_interact_chuckya,
    [id_bhvEnemyLakitu] = hitbox_interact_enemygeneral,
    [id_bhvSpiny] = hitbox_interact_spiny,
    [id_bhvWigglerHead] = hitbox_interact_wigglerhead,
    -- TTM
    [id_bhvMontyMole] = hitbox_interact_monty_mole,
    -- WDW
    [id_bhvSkeeter] = hitbox_interact_enemygeneral,
    -- SL
    [id_bhvSmallChillBully] = hitbox_interact_bully,
    [id_bhvBigChillBully] = hitbox_interact_bully,
    [id_bhvMoneybag] = hitbox_interact_enemygeneral,
    -- BOWSER
    [id_bhvBowser] = hitbox_interact_bowser,
    [id_bhvBowserBomb] = hitbox_interact_bowser_bomb,
}

local displacedhitboxes = {
    [id_bhvSnufit] = true,
    [id_bhvSwoop] = true
}

function hitboxattack(obj, playerIndex)
    local hitbox_dmg = obj.oHitboxDamage
    local hitbox_pos = { x = obj.oPosX, y = obj.oPosY, z = obj.oPosZ }
    local is_multi_hit = obj.oHitboxIsMultiHit
    local hitbox_type = obj.oHitboxType
    local hurtenemy = HITBOX_NO_HIT
    local enemyobj

    local actualtable = hitboxhurtableenemy
    local actualiterationstable = loadedinteractablebhvs
    -- Detect if SR enemies are present
    if (bhvSMSRShyGuy) then
        actualtable[bhvSMSRSmallBee] = hitbox_interact_enemygeneral
        actualtable[bhvSMSRShyGuy] = hitbox_interact_enemygeneral
        actualtable[bhvSMSRBreakableFloor] = hitbox_interact_srbreakable
        actualtable[bhvSMSRBreakableWindow] = hitbox_interact_srbreakable
        actualtable[bhvSMSRBreakableRock] = hitbox_interact_srbreakable
        actualtable[bhvSMSRWigglerHead] = hitbox_interact_wigglerhead
        actualtable[bhvSMSRAttractedSpaceBox] = hitbox_interact_srexplosion
        actualtable[bhvSMSRYoshi] = hitbox_interact_srgeneral_npc
        actualtable[bhvSMSRPeachMessage] = hitbox_interact_srgeneral_npc

        actualiterationstable[bhvSMSRSmallBee] = true
        actualiterationstable[bhvSMSRShyGuy] = true
        actualiterationstable[bhvSMSRBreakableFloor] = true
        actualiterationstable[bhvSMSRBreakableWindow] = true
        actualiterationstable[bhvSMSRBreakableRock] = true
        actualiterationstable[bhvSMSRWigglerHead] = true
        actualiterationstable[bhvSMSRAttractedSpaceBox] = true
        actualiterationstable[bhvSMSRYoshi] = true
        actualiterationstable[bhvSMSRPeachMessage] = true
    end

    local mmx = gMmxStates[playerIndex]
    
    for key, _ in pairs(actualiterationstable) do --actualtable is a table of enemies that can be hurt by hitboxes with each enemy stored by behaviorid
        local displacedOffset = 0
        if displacedhitboxes[key] then
            displacedOffset = 50
        end
        enemyobj = obj_get_nearest_object_with_behavior_id(obj,key) --get nearest hitable object
        if enemyobj then
            if HITBOX_GROUP_MELEE[hitbox_type] then
                for i = obj.oFurthestHitbox, 1, -1 do
                    local hitboxIndex = i
                    local hitboxX = obj.parentObj.oHitboxPosX + sins(obj.parentObj.oHitboxYaw) * coss(obj.parentObj.oHitboxPitch) * obj.hitboxRadius * hitboxIndex
                    local hitboxY = obj.parentObj.oHitboxPosY - sins(obj.parentObj.oHitboxPitch) * (obj.hitboxHeight - obj.hitboxDownOffset) * hitboxIndex + obj.hitboxHeight / 2 - obj.hitboxDownOffset
                    local hitboxZ =  obj.parentObj.oHitboxPosZ + coss(obj.parentObj.oHitboxYaw) * coss(obj.parentObj.oHitboxPitch) * obj.hitboxRadius * hitboxIndex
                    if obj_check_overlap_with_hitbox_params(enemyobj, hitboxX, hitboxY, hitboxZ, obj.hitboxHeight, obj.hitboxRadius, obj.hitboxDownOffset - displacedOffset) then
                        if actualtable[key] and (is_multi_hit == 1 or (mmx.active_hitboxes[obj.oHitboxGlobalIndex] ~= nil and not mmx.active_hitboxes[obj.oHitboxGlobalIndex][enemyobj.oSyncID])) then
                            set_health_for_overridden(enemyobj, key)
                            local result = actualtable[key](enemyobj, obj.oHitboxGlobalIndex, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, obj.oMoveAngleYaw)
                            if result > hurtenemy then
                                hurtenemy = result
                            end
                            break
                        end
                    end
                end
            else
                if ((obj_check_hitbox_overlap(obj, enemyobj) and not displacedhitboxes[key]) 
                    or (obj_check_overlap_with_hitbox_params(obj, enemyobj.oPosX, enemyobj.oPosY, enemyobj.oPosZ, enemyobj.hitboxHeight, enemyobj.hitboxRadius, 50) and displacedhitboxes[key])) then
                    if actualtable[key] and (is_multi_hit == 1 or (mmx.active_hitboxes[obj.oHitboxGlobalIndex] ~= nil and not mmx.active_hitboxes[obj.oHitboxGlobalIndex][enemyobj.oSyncID])) then
                        set_health_for_overridden(enemyobj, key)
                        local result = actualtable[key](enemyobj, obj.oHitboxGlobalIndex, hitbox_dmg, hitbox_pos, playerIndex, hitbox_type, obj.oMoveAngleYaw)
                        if result > hurtenemy then
                            hurtenemy = result
                        end
                    end
                end
            end
        end
    end
    return hurtenemy
end

function hitbox_interact_other_players(obj)
    local np = network_player_from_global_index(obj.oOwner)
    local m = gMarioStates[np.localIndex]
    local player = nearest_mario_state_to_object(obj)
    if player ~= nil and obj_check_hitbox_overlap(obj, player.marioObj) and player.playerIndex ~= m.playerIndex then
        if (obj.oHitboxInteractsWithOthers == 0) then
            return true
        end
    end
    return false
end


-----------------------------------------------------------------
-----------------------------------------------------------------
----- HITBOX BEHAVIOR -------------------------------------------
-----------------------------------------------------------------
-----------------------------------------------------------------

function hitbox_init(o)
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    if m.playerIndex == 0 then
        local s = gPlayerSyncTable[0]
        s.mmx_hit_stop = 0
    end
    o.oHitboxInteractsWithOthers = 0
    o.oHitboxCanHitTimer = 0
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj_set_model_extended(o, E_model_hitcylinder)
    cur_obj_hide()
    obj_scale_xyz(o, o.hitboxRadius / 100, o.hitboxHeight / 100, o.hitboxRadius / 100)
    if gServerSettings.playerInteractions ~= PLAYER_INTERACTIONS_NONE then
        if o.oHitboxType ~= HITBOX_SHOT_WEAK and o.oHitboxType ~= HITBOX_SHOT_MEDIUM then
            o.oHitboxInteractsWithOthers = 1
            if (m.playerIndex ~= 0) then
                o.oInteractType = INTERACT_DAMAGE
            end
        end
        o.oIntangibleTimer = 0
        if gServerSettings.playerInteractions == PLAYER_INTERACTIONS_PVP then
            o.oHitboxInteractsWithOthers = 1
            if (m.playerIndex ~= 0) then
                o.oInteractType = INTERACT_DAMAGE
                if o.oHitboxType == HITBOX_SHOT_WEAK then
                    o.oDamageOrCoinValue = 0
                elseif o.oHitboxType == HITBOX_SHOT_MEDIUM then
                    o.oDamageOrCoinValue = 1
                elseif HITBOX_GROUP_BURST[o.oHitboxType] then
                    o.oDamageOrCoinValue = 3
                else
                    o.oDamageOrCoinValue = 2
                end
            end
        end
    end
end

function hitbox_loop(o)
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local s = gPlayerSyncTable[m.playerIndex]
    if s.mmx_hit_stop > 0 then
        o.oTimer = o.oTimer - 1
    end

    if (HITBOX_GROUP_MELEE[o.oHitboxType]) then
        if (o.oTimer > 0) then
            if (o.oHitboxPositionIndex == o.oFurthestHitbox) then
                if o.oHitboxCanHitTimer == 0 and hitboxattack(o, m.playerIndex) > HITBOX_NO_HIT then
                    if (o.oHitboxHitstop > 0) then
                        if m.playerIndex == 0 and s.mmx_hit_stop == 0 then
                            s.mmx_hit_stop = o.oHitboxHitstop
                        end
                    end
                    o.oHitboxCanHitTimer = o.oHitboxWaitTimesBetweenHits
                end
            end
        end
    elseif (HITBOX_GROUP_PROJECTILE[o.oHitboxType]) then
        if o.oHitboxDestroyNextFrame == 1 then
            obj_mark_for_deletion(o)
        end
        local hitboxStep = hitboxattack(o, m.playerIndex)
        if hitboxStep ~= HITBOX_NO_HIT then
            if o.oHitboxType == HITBOX_SHOT_WEAK or (o.oHitboxType ~= HITBOX_SHOT_WEAK and hitboxStep == HITBOX_HIT_NO_DEATH) then
                o.oHitboxDestroyNextFrame = 1
            end
        end
    elseif (HITBOX_GROUP_BURST[o.oHitboxType]) then
        if o.oHitboxCanHitTimer == 0 and hitboxattack(o, m.playerIndex) then
            o.oHitboxCanHitTimer = o.oHitboxWaitTimesBetweenHits
        end
    end
    
    local destroyEarlierByFrames = 0
    if o.oHitboxType == HITBOX_MELEE_HEAVY_SLASH then
        destroyEarlierByFrames = 1
        if o.oTimer > 5 then
            o.hitboxRadius = 37.5 - (o.oTimer - 6) * 7
        end
    end
    
    if HITBOX_GROUP_MELEE[o.oHitboxType] then
        o.oPosX = o.parentObj.oHitboxPosX + sins(o.parentObj.oHitboxYaw) * coss(o.parentObj.oHitboxPitch) * o.hitboxRadius * o.oHitboxPositionIndex
        o.oPosY = o.parentObj.oHitboxPosY - sins(o.parentObj.oHitboxPitch) * (o.hitboxHeight - o.hitboxDownOffset) * o.oHitboxPositionIndex + o.hitboxHeight / 2 - o.hitboxDownOffset
        o.oPosZ = o.parentObj.oHitboxPosZ + coss(o.parentObj.oHitboxYaw) * coss(o.parentObj.oHitboxPitch) * o.hitboxRadius * o.oHitboxPositionIndex
    else
        o.oPosX = o.parentObj.oHitboxPosX
        o.oPosY = o.parentObj.oHitboxPosY + o.hitboxHeight / 2 - o.hitboxDownOffset
        o.oPosZ = o.parentObj.oHitboxPosZ
    end

    -- Hit stop
    if (o.oHitboxPositionIndex == o.oFurthestHitbox and HITBOX_GROUP_MELEE[o.oHitboxType] and o.oHitboxHitstop > 0) then
        if (m.playerIndex == 0) then
            if (s.mmx_hit_stop == o.oHitboxHitstop) then
                enable_time_stop_including_mario()
            end
    
            if s.mmx_hit_stop > 0 then
                s.mmx_hit_stop = s.mmx_hit_stop - 1
            end
            
            if (s.mmx_hit_stop == 0) then
                if o.activeFlags & ACTIVE_FLAG_INITIATED_TIME_STOP ~= 0 then
                    disable_time_stop_including_mario()
                end
            end
        end
    end

    if (o.oHitboxCanHitTimer ~= nil and o.oHitboxCanHitTimer > 0) then
        o.oHitboxCanHitTimer = o.oHitboxCanHitTimer - 1
    end

    if (HITBOX_GROUP_MELEE[o.oHitboxType]) then
        if (zSaberDeleteConditionTable[o.oHitboxType](m, o.oTimer + destroyEarlierByFrames) ~= false
         or obj_has_parent_bhv(o, hitbox_type_parent[o.oHitboxType]) == false) then
            if (o.oHitboxPositionIndex == o.oFurthestHitbox and o.oHitboxHitstop > 0) then
                o.oHitboxCanHitTimer = 0
                if m.playerIndex == 0 then
                    s.mmx_hit_stop = 0
                    if o.activeFlags & ACTIVE_FLAG_INITIATED_TIME_STOP ~= 0 then
                        disable_time_stop_including_mario()
                    end
                end
            end
            obj_mark_for_deletion(o)
        end
    else
        if obj_has_parent_bhv(o, hitbox_type_parent[o.oHitboxType]) == false then
            obj_mark_for_deletion(o)
        end
    end

    if hitbox_interact_other_players(o) then
        if HITBOX_GROUP_PROJECTILE[o.oHitboxType] then
            o.oHitboxDestroyNextFrame = 1
        end
    end

    local mmx = gMmxStates[m.playerIndex]
    if mmx.active_hitboxes[o.oHitboxGlobalIndex] == nil then
        obj_mark_for_deletion(o)
    end
end

id_bhvHitbox = hook_behavior(nil, OBJ_LIST_GENACTOR, false, hitbox_init, hitbox_loop)

-----------------------------------------------------------------
-----------------------------------------------------------------
----- HOOKS -----------------------------------------------------
-----------------------------------------------------------------
-----------------------------------------------------------------

local function on_delete_hitbox(m, o)
    local np = network_player_from_global_index(o.oOwner)
    local mOwner = gMarioStates[np.localIndex]
    local mmxOwner = gMmxStates[mOwner.playerIndex]
    if m.playerIndex == 0 and mmxOwner.active_hitboxes[o.oHitboxGlobalIndex] then
        mmxOwner.active_hitboxes[o.oHitboxGlobalIndex] = nil
        network_send(true, { type = 'hitbox_id_sync', id = o.oHitboxGlobalIndex, m = o.oOwner, remove = 1})
    end
end

-- If a player interacts with someone else's hitbox, delete it
hook_event(HOOK_ON_INTERACT, function(m, o, intType)
    if (o ~= nil and get_id_from_behavior(o.behavior) == id_bhvHitbox and intType == INTERACT_DAMAGE) then     
        if _G.HideAndSeek then
            local np = network_player_from_global_index(o.oOwner)
            local ownerMarioState = gMarioStates[np.localIndex]
            if _G.HideAndSeek.is_player_seeker(ownerMarioState.playerIndex) and not _G.HideAndSeek.is_player_seeker(m.playerIndex) then
                _G.HideAndSeek.set_player_seeker(m.playerIndex, true)
            end
        end
        if HITBOX_GROUP_PROJECTILE[o.oHitboxType] then
            on_delete_hitbox(m, o)
        end
    end
end)

hook_event(HOOK_ON_OBJECT_UNLOAD, function(o)
    if (o ~= nil and get_id_from_behavior(o.behavior) == id_bhvHitbox) then
        local np = network_player_from_global_index(o.oOwner)
        local m = gMarioStates[np.localIndex]
        local s = gPlayerSyncTable[m.playerIndex]
        if (m.playerIndex == 0) then
            if o.oHitboxType == HITBOX_SHOT_WEAK then
                s.x_active_shots = s.x_active_shots - 1
            end
            on_delete_hitbox(m, o)
        end
    end
end)

-- hitbox collision iterations optimizations

hook_event(HOOK_ON_LEVEL_INIT, function()
    map_hitbox_type_to_valid_parents()
    loadedinteractablebhvs = {}
    for key, _ in pairs(hitboxhurtableenemy) do
        local obj = obj_get_first_with_behavior_id(key)
        if obj then
            loadedinteractablebhvs[get_id_from_behavior(obj.behavior)] = true
        end
    end
end)

hook_event(HOOK_ON_OBJECT_LOAD, function(o)
    if hitboxhurtableenemy[get_id_from_behavior(o.behavior)] ~= nil then
        loadedinteractablebhvs[get_id_from_behavior(o.behavior)] = true
    end
end)

hook_event(HOOK_ON_OBJECT_UNLOAD, function(o)
    if not obj_get_first_with_behavior_id(get_id_from_behavior(o.behavior)) and not obj_get_next_with_same_behavior_id(o) then
        loadedinteractablebhvs[get_id_from_behavior(o.behavior)] = nil
    end
end)