-----------------------------------------------------------------
----- LOCALIZED FUNCTIONS ---------------------------------------
-----------------------------------------------------------------
local math_sin, math_cos, math_pi, sins, coss, math_random, math_pi, network_player_from_global_index, obj_mark_for_deletion, obj_spawn_loot_yellow_coins,
 network_send_object, mmx_play_sound, obj_has_behavior_id, calculate_yaw, network_init_object, obj_check_overlap_with_hitbox_params,
 nearest_mario_state_to_object, obj_get_nearest_object_with_behavior_id, obj_check_hitbox_overlap, obj_scale, collision_find_surface_on_ray =
 math.sin, math.cos, math.pi, sins, coss, math.random, math.pi, network_player_from_global_index, obj_mark_for_deletion, obj_spawn_loot_yellow_coins,
  network_send_object, mmx_play_sound, obj_has_behavior_id, calculate_yaw, network_init_object, obj_check_overlap_with_hitbox_params,
  nearest_mario_state_to_object, obj_get_nearest_object_with_behavior_id, obj_check_hitbox_overlap, obj_scale, collision_find_surface_on_ray


SHOT_TYPE_X_DEFAULT = 1000
SHOT_TYPE_X_LIGHT_SMALL = 2000
SHOT_TYPE_X_GIGA_MAIN = 3000
SHOT_TYPE_X_GIGA_SMALL = 4000
SHOT_TYPE_X_GAEA = 5000
SHOT_TYPE_VILE_DEFAULT = 6000

local function normal_busterOffset(faceAngle, armTheta, armLength)
    return {
        x = sins(faceAngle + armTheta) * armLength,
        z = coss(faceAngle + armTheta) * armLength
    }
end

local function light_last_busterOffset(faceAngle, armTheta, armLength)
    return {
        x = sins(faceAngle + armTheta) * armLength - sins(faceAngle + 16384) * 50,
        z = coss(faceAngle + armTheta) * armLength - coss(faceAngle + 16384) * 50
    }
end

local function giga_small_busterOffset(faceAngle, armTheta, armLength)
    return {
        x = sins(faceAngle + armTheta + 8000) * armLength,
        z = coss(faceAngle + armTheta + 8000) * armLength
    }
end

SOUND_X_SHOT1_IDX = 6
SOUND_X_SHOT2_IDX = 7
SOUND_X_SHOT3_IDX = 8
SOUND_X_SHOT_MAX_IDX = 9
SOUND_VILE_SHOT_IDX = 10
SOUND_VILE_CANNON_IDX = 11

local soundsTable = {
    [SOUND_X_SHOT1_IDX] = SOUND_X_SHOT1,
    [SOUND_X_SHOT2_IDX] = SOUND_X_SHOT2,
    [SOUND_X_SHOT3_IDX] = SOUND_X_SHOT3,
    [SOUND_X_SHOT_MAX_IDX] = SOUND_X_SHOT_MAX,
    [SOUND_VILE_SHOT_IDX] = SOUND_VILE_SHOT,
    [SOUND_VILE_CANNON_IDX] = SOUND_VILE_CANNON
}

-- "a" stands for armor
-- "c" stands for charge level
local spawnShotDataTable = {
    ['x-a0-c0'] = {
        [1] = {oHitboxDamage = 0.95, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_WEAK, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT1_IDX, soundVolume = 0.45},
    },
    ['x-a0-c1'] = {
        [1] = {oHitboxDamage = 3.8, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_MEDIUM, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT2_IDX, soundVolume = 0.73},
    },
    ['x-a0-c2'] = {
        [1] = {oHitboxDamage = 8.6, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT3_IDX, soundVolume = 0.73},
    },
    ['x-a1-c0'] = {
        [1] = {oHitboxDamage = 0.95, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_WEAK, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT1_IDX, soundVolume = 0.45},
    },
    ['x-a1-c1'] = {
        [1] = {oHitboxDamage = 3.8, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_MEDIUM, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT2_IDX, soundVolume = 0.73},
    },
    ['x-a1-c2'] = {
        [1] = {oHitboxDamage = 7, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT3_IDX, soundVolume = 0.73},
    },
    ['x-a1-c3'] = {
        [1] = {oHitboxDamage = 6, oShotType = SHOT_TYPE_X_LIGHT_SMALL, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT_MAX_IDX, soundVolume = 0.85},
        [2] = {oHitboxDamage = 6, oShotType = SHOT_TYPE_X_LIGHT_SMALL, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = math_pi, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = -15},
        [3] = {oHitboxDamage = 6, oShotType = SHOT_TYPE_X_LIGHT_SMALL, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 1, horOffsetFunc = light_last_busterOffset, extraVerOffset = 0},
    },
    ['x-a2-c0'] = {
        [1] = {oHitboxDamage = 0.95, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_WEAK, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT1_IDX, soundVolume = 0.45},
    },
    ['x-a2-c1'] = {
        [1] = {oHitboxDamage = 3.8, oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_MEDIUM, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT2_IDX, soundVolume = 0.73},
    },
    ['x-a2-c2'] = {
        [1] = {oHitboxDamage = 7,oShotType = SHOT_TYPE_X_DEFAULT, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT3_IDX, soundVolume = 0.73},
    },
    ['x-a2-c3'] = {
        [1] = {oHitboxDamage = 8, oShotType = SHOT_TYPE_X_GIGA_MAIN, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT_MAX_IDX, soundVolume = 0.85},
        [2] = {oHitboxDamage = 5, oShotType = SHOT_TYPE_X_GIGA_SMALL, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = giga_small_busterOffset, extraVerOffset = -70},
    },
    ['x-a4-c0'] = {
        [1] = {oHitboxDamage = 0.95, oShotType = SHOT_TYPE_X_GAEA, oHitboxType = HITBOX_SHOT_WEAK, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT1_IDX, soundVolume = 0.45},
    },
    ['x-a4-c2'] = {
        [1] = {oHitboxDamage = 7, oShotType = SHOT_TYPE_X_GAEA, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_X_SHOT3_IDX, soundVolume = 0.73},
    },
    ['vile-c0'] = {
        [1] = {oHitboxDamage = 0.95, oShotType = SHOT_TYPE_VILE_DEFAULT, oHitboxType = HITBOX_SHOT_WEAK, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_VILE_SHOT_IDX, soundVolume = 0.44},
    },
    ['vile-c2'] = {
        [1] = {oHitboxDamage = 5.9, oShotType = SHOT_TYPE_VILE_DEFAULT, oHitboxType = HITBOX_SHOT_STRONG, oShotInitOffset = 0, 
        oShotInitRotPhase = 0, horOffsetFunc = normal_busterOffset, extraVerOffset = 0, sound = SOUND_VILE_CANNON_IDX, soundVolume = 0.54},
    },
}

-- Maybe faster than tostring?
local numberToString = {
    [0] = '0',
    [1] = '1',
    [2] = '2',
    [3] = '3',
    [4] = '4',
}

function spawn_shot(m, armHeight, armLength, armTheta, chargeLevel)
    local faceAngle = m.faceAngle.y
    if m.action == ACT_WALL_SLIDE then
        if m.wall then
            faceAngle = atan2s(m.wallNormal.z, m.wallNormal.x)
        else
            faceAngle = m.faceAngle.y + 32768
        end
    end
    local mGfx = m.marioObj.header.gfx
    local s = gPlayerSyncTable[m.playerIndex]

    local fullString = s.character
    if (s.character == 'x') then
        local armorToString = numberToString[s.x_armor]
        fullString = fullString .. '-a' .. armorToString
        if (m.playerIndex == 0 and chargeLevel > 0) then
            s.x_shoot_wait = 5
        end
    end
    local chargeLevelToString = numberToString[chargeLevel]
    fullString = fullString .. '-c' .. chargeLevelToString
    for key, _ in pairs(spawnShotDataTable[fullString]) do
        local data = spawnShotDataTable[fullString][key]
        if m.playerIndex == 0 then
            if data.sound then
                mmx_play_sound(soundsTable[data.sound], m.pos, data.soundVolume)
                network_send(true, {type = 'sound', sound_idx = data.sound, intensity = data.soundVolume, m = gNetworkPlayers[m.playerIndex].globalIndex})
            end
            local id = assign_id_to_hitbox(m)
            local offset = data.horOffsetFunc(faceAngle, armTheta, armLength)
            if data.oHitboxType == HITBOX_SHOT_WEAK then
                s.x_active_shots = s.x_active_shots + 1
            end
            spawn_non_sync_object(id_bhvShot, E_MODEL_NONE, mGfx.pos.x + offset.x, mGfx.pos.y + armHeight + data.extraVerOffset, mGfx.pos.z + offset.z, function(obj)
                obj.oMoveAngleYaw = faceAngle
                obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
                obj.oShotDestroyNextFrame = 0
                obj.oShotType = data.oShotType
                obj.oHitboxType = data.oHitboxType
                obj.oHitboxDamage = data.oHitboxDamage
                obj.oShotInitOffset = data.oShotInitOffset
                obj.oShotInitRotPhase = data.oShotInitRotPhase
                obj.oHitboxGlobalIndex = id
            end)
            network_send(true, {
                type = 'shot', 
                m = gNetworkPlayers[m.playerIndex].globalIndex,
                oMoveAngleYaw = faceAngle,
                oShotType = data.oShotType,
                oHitboxType = data.oHitboxType,
                oHitboxDamage = data.oHitboxDamage,
                oShotInitOffset = data.oShotInitOffset,
                oShotInitRotPhase = data.oShotInitRotPhase,
                posX = mGfx.pos.x + offset.x,
                posY = mGfx.pos.y + armHeight + data.extraVerOffset,
                posZ = mGfx.pos.z + offset.z,
                oHitboxGlobalIndex = id,
            })
        end
    end
end


-----------------------------------------------------------------
-----------------------------------------------------------------
----- BEHAVIORS -------------------------------------------------
-----------------------------------------------------------------
-----------------------------------------------------------------

local shotStats = {
    [HITBOX_SHOT_WEAK + SHOT_TYPE_X_DEFAULT] = {
        model = E_model_shot1,
        hitboxRadius  = 30,
        hitboxHeight = 68,
        hitboxDownOffset  = 52,
        oWallHitboxRadius = 20,
        oForwardVel = 92,
        shotLife = 31,
        scale = 1,
        initScale = 0.9,
        scalePulse = 0,
        shotRing = nil,
        trailModel = nil,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = false,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_MEDIUM + SHOT_TYPE_X_DEFAULT] = {
        model = E_model_shot2,
        hitboxRadius  = 70,
        hitboxHeight = 95,
        hitboxDownOffset  = 73,
        oWallHitboxRadius = 60,
        oForwardVel = 130,
        shotLife = 37,
        scale = 1.15,
        initScale = 1,
        scalePulse = 0,
        shotRing = nil,
        trailModel = nil,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = false,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_STRONG + SHOT_TYPE_X_DEFAULT] = {
        model = E_model_shot3,
        hitboxRadius  = 130,
        hitboxHeight = 160,
        hitboxDownOffset  = 110,
        oWallHitboxRadius = 120,
        oForwardVel = 170,
        shotLife = 45,
        scale = 1.4,
        initScale = 1,
        scalePulse = 1.3,
        shotRing = E_model_shot3_ring,
        trailModel = E_model_shot3_particle,
        trailHorOffset = 40,
        trailYOffset = 50,
        canRoll = true,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_STRONG + SHOT_TYPE_X_LIGHT_SMALL] = {
        model = E_model_shot_max,
        hitboxRadius  = 87,
        hitboxHeight = 105,
        hitboxDownOffset  = 80,
        oWallHitboxRadius = 77,
        oForwardVel = 170,
        shotLife = 50,
        scale = 1.4,
        initScale = 1,
        scalePulse = 0,
        shotRing = nil,
        trailModel = E_model_shot_max,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = false,
        trajectory = 'spiral',
        spiralRadius = 65,
    },
    [HITBOX_SHOT_STRONG + SHOT_TYPE_X_GIGA_MAIN] = {
        model = E_model_giga_pink_shot,
        hitboxRadius  = 130,
        hitboxHeight = 160,
        hitboxDownOffset  = 110,
        oWallHitboxRadius = 120,
        oForwardVel = 170,
        shotLife = 50,
        scale = 1.4,
        initScale = 1,
        scalePulse = 1.3,
        shotRing = E_model_shot3_ring,
        trailModel = E_model_shot_max,
        trailHorOffset = 40,
        trailYOffset = 50,
        canRoll = true,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_STRONG + SHOT_TYPE_X_GIGA_SMALL] = {
        model = E_model_shot3_particle,
        hitboxRadius  = 87,
        hitboxHeight = 105,
        hitboxDownOffset  = 80,
        oWallHitboxRadius = 77,
        oForwardVel = 170,
        shotLife = 50,
        scale = 1.2,
        initScale = 1,
        scalePulse = 0,
        shotRing = nil,
        trailModel = E_model_shot3_particle,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = false,
        trajectory = 'spiral',
        spiralRadius = 100,
    },
    [HITBOX_SHOT_WEAK + SHOT_TYPE_X_GAEA] = {
        model = E_model_shot_gaea_weak,
        hitboxRadius  = 34,
        hitboxHeight = 72,
        hitboxDownOffset  = 54,
        oWallHitboxRadius = 22,
        oForwardVel = 92,
        shotLife = 10,
        scale = 1,
        initScale = 0.8,
        scalePulse = 0,
        shotRing = nil,
        trailModel = nil,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = false,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_STRONG + SHOT_TYPE_X_GAEA] = {
        model = E_model_shot_gaea_charged,
        hitboxRadius  = 130,
        hitboxHeight = 160,
        hitboxDownOffset  = 110,
        oWallHitboxRadius = 120,
        oForwardVel = 170,
        shotLife = 7,
        scale = 1.4,
        initScale = 1,
        scalePulse = 1.3,
        shotRing = E_model_shot_gaea_ring,
        trailModel = E_model_shot_gaea_trail,
        trailHorOffset = 40,
        trailYOffset = 50,
        canRoll = true,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_WEAK + SHOT_TYPE_VILE_DEFAULT] = {
        model = E_model_vulcan_shot,
        hitboxRadius  = 30,
        hitboxHeight = 68,
        hitboxDownOffset  = 52,
        oWallHitboxRadius = 20,
        oForwardVel = 92,
        shotLife = 25,
        scale = 0.4,
        initScale = 0.3,
        scalePulse = 0,
        shotRing = nil,
        trailModel = nil,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = false,
        trajectory = 'linear',
        spiralRadius = 0,
    },
    [HITBOX_SHOT_STRONG + SHOT_TYPE_VILE_DEFAULT] = {
        model = E_model_cannon_shot,
        hitboxRadius  = 120,
        hitboxHeight = 150,
        hitboxDownOffset  = 100,
        oWallHitboxRadius = 110,
        oForwardVel = 140,
        shotLife = 45,
        scale = 1.4,
        initScale = 1,
        scalePulse = 1.3,
        shotRing = nil,
        trailModel = nil,
        trailHorOffset = 0,
        trailYOffset = 0,
        canRoll = true,
        trajectory = 'linear',
        spiralRadius = 0,
    },
}


local shot_init = function(o)
    o.oFlags = (OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_ANGLE_TO_MOVE_ANGLE | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW)
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local stats = shotStats[o.oHitboxType + o.oShotType]
    o.oForwardVel = stats.oForwardVel
    o.hitboxRadius = stats.hitboxRadius
    o.hitboxHeight = stats.hitboxHeight
    o.hitboxDownOffset = stats.hitboxDownOffset
    o.oWallHitboxRadius = stats.oWallHitboxRadius
    o.oWallHitboxRadius = stats.oWallHitboxRadius
    obj_scale(o, stats.initScale)
    obj_set_model_extended(o, stats.model)
    spawn_non_sync_object(id_bhvHitbox, E_MODEL_NONE, o.oPosX, o.oPosY, o.oPosZ, function(obj)
        obj.oOwner = o.oOwner
        obj.oHitboxType = o.oHitboxType
        obj.oShotType = o.oShotType
        obj.oHitboxGlobalIndex = o.oHitboxGlobalIndex
        obj.oHitboxDamage = o.oHitboxDamage
        obj.hitboxRadius = stats.hitboxRadius
        obj.hitboxHeight = stats.hitboxHeight
        obj.hitboxDownOffset = stats.hitboxDownOffset
        obj.oWallHitboxRadius = stats.oWallHitboxRadius
        obj.parentObj = o
    end)
    o.oShotDestroyNextFrame = 0
    if stats.shotRing ~= nil then
        spawn_non_sync_object(id_bhvShotRing, stats.shotRing, o.oPosX, o.oPosY, o.oPosZ, function(obj)
            obj.oForwardVel = 60
            obj.oMoveAngleYaw = o.oMoveAngleYaw
        end)
    end
    if o.oShotType == SHOT_TYPE_X_LIGHT_SMALL and o.oShotInitRotPhase == 1 then
        spawn_non_sync_object(id_bhvCharge2Front, E_model_shot_max_bubble, o.oPosX, o.oPosY, o.oPosZ, function(obj)
            obj.oForwardVel = 60
            obj.oMoveAngleYaw = o.oMoveAngleYaw
            obj.oOwner = gNetworkPlayers[m.playerIndex].globalIndex
            obj.oMoveAngleYaw = o.oMoveAngleYaw
        end)
    end
    o.oVelX = sins(o.oMoveAngleYaw) * o.oForwardVel
    o.oVelY = 0
    o.oVelZ = coss(o.oMoveAngleYaw) * o.oForwardVel
    set_hitbox_pos_and_angle_to_obj_pos_angle(o)
    o.hookRender = 1
end

local shot_loop = function(o)
    local np = network_player_from_global_index(o.oOwner)
    local m = gMarioStates[np.localIndex]
    local mmx = gMmxStates[m.playerIndex]
    local stats = shotStats[o.oHitboxType + o.oShotType]
    local shot_life = stats.shotLife
    o.hookRender = 1

    if o.oShotDestroyNextFrame == 1 then
        obj_mark_for_deletion(o)
    end

    -- Trails
    if stats.trailModel then
        local randomHorizontalOffset = 0
        local randomX = 0
        local randomZ = 0
        local randomY = 0
        if stats.trailHorOffset > 0 then
            randomHorizontalOffset = math_random(-0x8000, 0x8000)
            randomX = stats.trailHorOffset * sins(o.oMoveAngleYaw + randomHorizontalOffset)
            randomZ = stats.trailHorOffset * coss(o.oMoveAngleYaw + randomHorizontalOffset)
        end
        if stats.trailYOffset > 0 then
            randomY = math_random(-stats.trailYOffset, stats.trailYOffset)
        end
        spawn_non_sync_object(id_bhvShotTrailBall, stats.trailModel, o.oPosX + randomX, o.oPosY + randomY, o.oPosZ + randomZ, function(obj)
            obj.oForwardVel = 60
            obj.oMoveAngleYaw = o.oMoveAngleYaw
            obj.oShotType = o.oShotType
        end)
    end

    -- Roll
    if stats.canRoll then
        o.oMoveAngleRoll = o.oMoveAngleRoll + 4000 + 6000 / (o.oTimer * o.oTimer)
    end

    -- Scale
    if stats.scale then
        obj_scale(o, stats.scale)
    end

    if stats.scalePulse > 0 then
        if o.oTimer % 3 == 0 then
            o.header.gfx.scale.y = stats.scalePulse
        elseif o.oTimer % 2 == 0 then
            o.header.gfx.scale.x = stats.scalePulse
            o.header.gfx.scale.z = stats.scalePulse
        end
    end
    
    local collision = collision_find_surface_on_ray(
        o.oPosX, o.oPosY, o.oPosZ,
        o.oVelX, o.oVelY, o.oVelZ)

    if collision.surface ~= nil then
        o.oShotDestroyNextFrame = 1
    else
        if stats.trajectory == 'spiral' then
            local ang = (o.oMoveAngleYaw / 0x8000) * math_pi
            if o.oShotInitRotPhase == 0 then
                o.oPosX = o.oPosX + o.oVelX + stats.spiralRadius * math_sin(ang + o.oTimer - o.oShotInitOffset + math_pi / 2)
                o.oPosY = o.oPosY + o.oVelY + stats.spiralRadius* math_sin(o.oTimer)
                o.oPosZ = o.oPosZ + o.oVelZ + stats.spiralRadius * math_cos(ang + o.oTimer - o.oShotInitOffset + math_pi / 2)
            else
                o.oPosX = o.oPosX + o.oVelX + stats.spiralRadius * math_sin(ang + o.oTimer - o.oShotInitOffset)
                o.oPosY = o.oPosY + o.oVelY + stats.spiralRadius * math_cos(o.oTimer)
                o.oPosZ = o.oPosZ + o.oVelZ + stats.spiralRadius * math_cos(ang + o.oTimer - o.oShotInitOffset)
            end
        else
            o.oPosX = o.oPosX + o.oVelX
            o.oPosY = o.oPosY + o.oVelY
            o.oPosZ = o.oPosZ + o.oVelZ
        end
    end

    set_hitbox_pos_and_angle_to_obj_pos_angle(o)

    if o.oTimer >= shot_life or mmx.active_hitboxes[o.oHitboxGlobalIndex] == nil then
        obj_mark_for_deletion(o)
    end
end

----------------------------
--BEHAVIORS REGISTRY--------
----------------------------
id_bhvShot = hook_behavior(nil, OBJ_LIST_GENACTOR, false, shot_init, shot_loop)


--------------------------------------------------
--- RENDER ---------------------------------------
--------------------------------------------------

local shotModelsMatFix = {
    [E_model_shot2] = 2,
    [E_model_shot3] = 3,
    [E_model_shot_max] = 2,
    [E_model_giga_pink_shot] = 3
}

hook_event(HOOK_ON_OBJECT_RENDER, function(o)
    if o ~= nil and shotModelsMatFix[obj_get_model_id_extended(o)] then
        local mat = 0
        local dl = cast_graph_node(o.header.gfx.sharedChild.children)
        gfx_parse(dl.displayList, function(cmd, op)
            if op == G_SETCOMBINE then             
                if mat == shotModelsMatFix[obj_get_model_id_extended(o)] then
                    gfx_set_combine_lerp(cmd, 0, 0, 0, G_CCMUX_1, 0, 0, 0, G_CCMUX_ENVIRONMENT, 0, 0, 0, G_ACMUX_1, 0, 0, 0, G_ACMUX_ENVIRONMENT)
                else
                    gfx_set_combine_lerp(cmd, 0, 0, 0, G_CCMUX_PRIMITIVE, 0, 0, 0, G_CCMUX_ENVIRONMENT, 0, 0, 0, G_ACMUX_PRIMITIVE, 0, 0, 0, G_ACMUX_ENVIRONMENT)
                end
                mat = mat + 1   
            end   
        end)
    end
end)